/*********************************************************************/
/* tclparse V2.00 from 14.March 1997                                 */
/* Copyright 1996/97 Stephan Weiss, Heiko Grossmann                  */
/*                   University of Stuttgart, Germany                */
/* For details see the copyright file included in the distribution.  */
/* This notice must not be removed from the file.                    */
/*                                                                   */
/* routines for parsing tcl commands                                 */
/*                                                                   */
/*********************************************************************/

#include "../include/const.h"
#include "../include/file.h"
#include "../include/handler.h"
#include "../include/parsing.h"
#include "../include/display.h"
#include <string.h>
#include <iostream.h>


void ParseTclAfterCmd( void )
{
  //Das erste Argument einlesen.
  if( !ParseUnit() )
    return;

  if( !strcmp( Actual( unit ), "cancel" ) )
    ParseGeneric( 1, -1 );

  else if( !strcmp( Actual( unit ), "idle" ) )
    ParseGeneric( 0, -1 );

  else if( !strcmp( Actual( unit ), "info" ) )
    ParseGeneric( 0, 1 );

  else
    ParseGeneric( 1, -1 );
}


void ParseTclAppendCmd( void )
{
  //Den Variablennamen einlesen.
  if( !ParseUnit( "varName" ) )
    return;

  //Folgen keine Argumente mehr, so muss die Variable definiert sein.
  if( IsLastArg() )
    {
      if( !IsVariable( Actual( unit ) ) )
	display.UnknownVariable( Actual( unit ) );

      return;
    }

  char copy[ strlen( Actual( unit ) )+1 ];
  strcpy( copy, Buffer() );

  //Sonst die Variable im Programm aufnehmen.
  AddVariable( copy, Environment(), 1, 0 );

  //Die restlichen Argumente einlesen.
  ParseGeneric( 1, -1 );
}


void ParseTclArrayCmd( void )
{
  //Das erste Argument einlesen.
  if( !ParseUnit() )
    return;

  //Sich den Namen merken.
  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Buffer() );

  //Da bei jedem 'array'-Befehl als zweites Argument der Feldname erwartet
  //wird, kann man hier zentral diesen Namen fuer alle Befehle einlesen.
  if( !ParseUnit() )
    return;

  //Die Befehle 'anymore', 'donesearch' und 'nextelement' sind syntaktisch gleich.
  if( !strcmp( "anymore", command ) || !strcmp( "donesearch", command ) || !strcmp( "nextelement", command ) )
    {
      //Das Feld muss existieren.
      if( !IsArray( Actual( unit ) ) )
	display.Error( "variable '", Actual( unit ), "' is not an array" );

      //Das Argument 'searchID' einlesen.
      if( !ParseUnit( "searchID" ) )
	return;

      //Befehlsende einlesen.
      ParseCommandEnd();
    }

  else if( !strcmp( command, "exists" ) )
    ParseCommandEnd();

  //Die Befehle 'get' und 'names' sind syntaktisch gleich.
  else if( !strcmp( "get", command ) || !strcmp( "names", command ) )
    {
      //Das Argument 'pattern' ist optional.
      if( !ParseOptionalUnit() )
	return;

      else
	ParseCommandEnd();
    }

  else if( !strcmp( command, "set" ) )
    {
      //Die erwartete Liste einlesen.
      if( !ParseUnit() )
	return;

      //Normalerweise muessten die Feldelemente eventuell fuer das
      //Programm sichtbar gemacht werden.

      //Befehlsende einlesen.
      ParseCommandEnd();
    }

  else if( !strcmp( command, "size" ) )
    ParseCommandEnd();

  else if( !strcmp( command, "startsearch" ) )
    {
      //Das Feld muss existieren.
      if( !IsArray( Actual( unit ) ) )
	display.Error( "variable '", Actual( unit ), "' is not an array" );

      //Befehlsende einlesen.
      ParseCommandEnd();
    }

  //Fehler: Es wurde kein gueltiger Befehl angegeben.
  else
    display.UnknownOption( command, "array" );
}


void ParseTclBreakCmd( void )
{
  if( !IsLoopActive() )
    display.Warning( "used 'break' outside of a loop" );

  ParseCommandEnd();
}


void ParseTclCaseCmd( void )
{
  display.Warning( "'case' is not supported by this program, use switch" );
  HandleError();
}


void ParseTclCatchCmd( void )
{
  if( !ParseScript() )
    return;

  if( !ParseOptionalUnit() )
    return;

  char copy[ strlen( Actual( unit ) )+1 ];
  strcpy( copy, Buffer() );

  AddVariable( copy, Environment(), 1, 0 );
}


void ParseTclCdCmd( void )
{
  ParseGeneric( 0, 1 );
}


void ParseTclClockCmd( void )
{
  //Das erste Argument einlesen.
  if( !ParseUnit() )
    return;

  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Buffer() );

  //Den Unterbefehl bearbeiten.
  if( !strcmp( command, "clicks" ) )
    ParseCommandEnd();

  else if( !strcmp( command, "format" ) )
    {
      //Das Argument 'clockValue' einlesen.
      if( !ParseUnit( "clockValue" ) )
	return;

      if( !ParseOptions( "clock_format" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( command, "scan" ) )
    {
      //Das Argument 'dateString' einlesen.
      if( !ParseUnit( "dateString" ) )
	return;

      if( !ParseOptions( "clock_scan" ) )
	return;

      ParseCommandEnd();
    }

  else if( !strcmp( command, "seconds" ) )
    ParseCommandEnd();

  else
    display.UnknownOption( command, "clock" );
}


void ParseTclCloseCmd( void )
{
  ParseGeneric( 1, 1 );
}


void ParseTclConcatCmd( void )
{
  ParseGeneric( 0, -1 );
}


void ParseTclContinueCmd( void )
{
  if( !IsLoopActive() )
    display.Error( "used 'continue' outside of a loop" );

  ParseCommandEnd();
}


void ParseTclEofCmd( void )
{
  ParseGeneric( 1, 1 );
}


void ParseTclErrorCmd( void )
{
  //Das erste Argument einlesen.
  if( !ParseUnit( "message" ) )
    return;

  //Das optionale Argument 'info' einlesen.
  if( !ParseOptionalUnit() )
    return;

  //Wurde 'info' angegeben, so wird die globale Tcl-Variable 'errorInfo' normalerweise
  //damit initialisiert.
  CVariable *variable = GetVariable( "errorInfo", " " );
  variable->SetStatus( 1 );

  //Das optionale Argument 'code' einlesen.
  if( !ParseOptionalUnit() )
    return;

  //Wurde 'code' angegeben, so wird die globale Tcl-Variable 'errorCode' normalerweise
  //damit initialisiert.
  CVariable *variable2 = GetVariable( "errorCode", " " );
  variable2->SetStatus( 1 );

  ParseCommandEnd();
}


void ParseTclEvalCmd( void )
{
  ParseGeneric( 1, -1 );
}


void ParseTclExecCmd( void )
{
  if( !ParseOptions( "exec" ) )
    return;

  ParseGeneric( 1, -1 );
}


void ParseTclExitCmd( void )
{
  ParseGeneric( 0, -1 );
}


void ParseTclExprCmd( void )
{
  ParseGeneric( 1, -1 );
}


void ParseTclFblockedCmd( void )
{
  ParseGeneric( 1, 1 );
}


void ParseTclFconfigureCmd( void )
{
  //Das erste Argument 'channelID' einlesen.
  if( !ParseUnit( "channelID" ) )
    return;

  if( !ParseOptions( "fconfigure" ) )
    return;

  ParseCommandEnd();
}


void ParseTclFileCmd( void )
{
  //Das erste Argument einlesen.
  if( !ParseUnit() )
    return;

  //Zum Suchen innerhalb eines Strings vor und nach dem Unterbefehlsnamen ein
  //Leerzeichen einfuegen.
  char subCommand[ strlen( Actual( unit ) ) +1 ];
  strcpy( subCommand, Buffer() );

  //Zuerst schauen, ob der Unterbefehl einer der vielen mit einem Argument ist.
  if( IsElementOf( subCommand, "atime dirname executable exists extension isdirectory isfile mtime owned pathtype readable readlink rootname size split tail type writable " ) )
    ParseGeneric( 1, 1 );

  else if( IsElementOf( subCommand, " copy rename " ) )
    {
      if( !ParseOptions( "file_copy" ) )
	return;

      ParseGeneric( 2, -1 );
    }

  else if( !strcmp( subCommand, "delete" ) )
    {
      if( !ParseOptions( "file_delete" ) )
	return;

      ParseGeneric( 1, -1 );
    }

  else if( !strcmp( subCommand, "join" ) )
    ParseGeneric( 1, -1 );

  else if( IsElementOf( subCommand, " lstat stat " ) )
    {
      //Das Argument 'name' einlesen.
      if( !ParseUnit( "name" ) )
	return;

      //Den Variablennamen einlesen. Dabei handelt es sich um ein
      //Feld, bei dem mehrere Elemente gesetzt werden.
      if( !ParseUnit( "varName" ) )
	return;

      //Die Indexnamen in einem String speichern.
      char indexList[] = "atime ctime dev gid ino mode ntime nlink size type uid";

      //Den ersten Indexnamen aus dem String heraustrennen.
      char *dummy = strtok( indexList," " );

      char completeName[ wordLength ];

      //Alle Elemente des Feldes definieren.
      while( dummy != NULL )
	{
	  //Den vollstaendigen Variablennamen bilden.
	  strcpy( completeName, Actual( unit ) );
	  strcat( completeName, "(" );
	  strcat( completeName, dummy );
	  strcat( completeName, ")" );
	  
	  //Die Variable fuer das Programm definieren.
	  if( AddVariable( completeName, " ", 1 ) )
	      display.NewVariable( completeName );

	  //Den naechsten Indexnamen ermitteln.
	  dummy = strtok( NULL," " );
	}
    }

  else if( !strcmp( subCommand, "mkdir" ) )
    ParseGeneric( 1, -1 );

  else
    display.UnknownOption( subCommand, "file" );
}


void ParseTclFileeventCmd( void )
{
  //Das Argument 'channelID' einlesen.
  if( !ParseUnit( "channelID" ) )
    return;

  //Den Namen des Unterbefehls ('readable' oder 'writable') einlesen.
  if( !ParseUnit( "readable' or 'writable" ) )
    return;

  //Schauen, ob wirklich 'readable' oder 'writable' eingelesen wurde.
  if( strcmp( "readable", Actual( unit ) ) || strcmp( "readable", Actual( unit ) ) )
    {
      display.Error( "'readable' or 'writable' expected" );
      return;
    }

  //Eventuell wurde noch ein Skript angegeben.
  if( IsLastArg() )
    return;

  if( !ParseScript() )
    return;

  ParseCommandEnd();
}


void ParseTclFlushCmd( void )
{
  ParseGeneric( 1, 1 );
}


void ParseTclForCmd( void )
{
  //Zuerst den 'start'-Rumpf einlesen.
  if( !ParseScript() )
    return;

  //Dann den 'test'-Ausdruck.
  if( !ParseUnit( "test" ) )
    return;

  //Den 'next'-Rumpf einlesen.
  if( !ParseScript() )
    return;

  //Den eigentlichen 'for'-Rumpf einlesen.
  if( !ParseScript() )
    return;
}


void ParseTclForeachCmd( void )
{
  if( IsLastArg() )
    {
      display.Error( "argument 'start' expected" );
      return;
    }

  //Zuerst die Anzahl der Argumente ermitteln;

  //Den aktuellen Zustand speichern.
  actualFile->EnableScanFlag();
  actualFile->Save();

  int count = 0;

  //Die Argumente einlesen und zaehlen.
  while( ParseOptionalUnit() )
    count++;

  //Den vorherigen Zustand wiederherstellen.
  actualFile->Restore();
  actualFile->DisableScanFlag();

  //Die ermittelte Zahl auswerten.
  //Die Anzahl der Argumente muss ungerade und mindestens 3 sein.
  if( (count % 2) == 0 )
    {
      display.Error( "even number of arguments not possible" );
      HandleError();
      return;
    }

  else if( count < 3 )
    {
      display.Error( "too few arguments" );
      HandleError();
      return;
    }

  char *dummy;
  char list[ bufferSize ];
  CCommand *procedure;
  CVariable variable;

  //Zuerst die Liste mit den verwendeten Variablen und den jeweils zugehoerigen
  //Wertelisten einlesen.
  for( int i=1 ; i < count ; i += 2 )
    {
      //Nicht 100% korrekt, aber wir verwenden 'ParseArgumentList'.
      procedure = ParseArgumentList();
      strcpy( list, procedure->List() );

      //Jede ermittelte Variable sichtbar machen.
      if( procedure->Min() > 0 )
	{
	  char *dummy = strtok( list, " " );
	  while( dummy != NULL )
	    {
	      AddVariable( dummy, Environment(), 1, 0 );
	      dummy = strtok( NULL, " " );
	    }
	}

      //Die zugehoerige Werteliste auslesen.
      GetUnit();
    }

  //Als letztes den Rumpf auswerten;
  if( !ParseScript() )
    return;

  ParseCommandEnd();
}


void ParseTclFormatCmd( void )
{
  ParseGeneric( 1, -1 );
}


void ParseTclGetsCmd( void )
{
  if( !ParseUnit() )
    return;

  //Danach den optionalen Variablennamen einlesen.
  if( !ParseOptionalUnit() )
    return;

  char copy[ strlen( Actual( unit ) )+1 ];
  strcpy( copy, Buffer() );

  //Wurde der Name angegeben, so wird die Variable fuer das Programm definiert.
  AddVariable( copy, Environment(), 1, 0 );

  ParseCommandEnd();
}


void ParseTclGlobCmd( void )
{
  if( !ParseOptions( "glob" ) )
    return;

  ParseGeneric( 1, -1 );
}


void ParseTclGlobalCmd( void )
{
  //'global' wird nur innerhalb von Prozeduren ausgefuehrt.
  if( !strcmp( Environment(), " " ) )
    {
      display.Warning( "command 'global' used outside of a procedure definition" );
      HandleError();
      return;
    }

  if( IsLastArg() )
    {
      display.Error( "variable expected" );
      return;
    }

  while( ParseOptionalUnit() )
    AddVariable( Actual( unit ), " ", 0, 0 );
}


void ParseTclHistoryCmd( void )
{
  //Den Namen des Unterbefehls einlesen.
  if( !ParseUnit() )
    return;

  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Buffer() );

  if( !strcmp( "add", command ) )
    {
      if( !ParseUnit() )
	return;

      if( !ParseOptionalUnit() )
	return;

      if( strcmp( "exec", Actual( unit ) ) )
	display.Error( "'exec' expected" );

      ParseCommandEnd();
    }

  else if( !strcmp( "change", command ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( "event", command ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( "info", command ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( "keep", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "nextID", command ) )
    ParseCommandEnd();

  else if( !strcmp( "redo", command ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( "substitute", command ) )
    ParseGeneric( 2, 3 );

  else if( !strcmp( "words", command ) )
    ParseGeneric( 1, 2 );

  else
    display.UnknownOption( command, "history" );
}


void ParseTclIfCmd( void )
{
  //Den 'test'-Ausdruck einlesen.
  if( !ParseUnit( "expr1" ) )
    return;

  //Optionales 'then' einlesen.
  if( !strcmp( "then", Next( word ) ) )
    GetUnit();

  //Den zugehoerigen Rumpf bearbeiten.
  if( !ParseScript() )
    return;

  //Hier kann schon das Befehlsende erreicht sein.
  if( IsLastArg() )
    return;

  //Sonst in einer Schleife saemtliche 'elseif'-Aeste einlesen.
  while( 1 )
    {
      if( strcmp( "elseif", Next( word ) ) )
	break;

      else
	GetUnit();

      //Den 'test'-Ausdruck einlesen.
      if( !ParseUnit( "expr" ) )
	return;

      //Optionales 'then' einlesen.
      if( !strcmp( "then", Next( word ) ) )
	GetUnit();

      //Den zugehoerigen Rumpf bearbeiten.
      if( !ParseScript() )
	return;

      //Eventuelles Befehlsende abfragen.
      if( IsLastArg() )
	return;
    }

  //Optionales 'else' einlesen.
  if( !strcmp( "else", Next( word ) ) )
    GetUnit();

  //Den zugehoerigen Rumpf bearbeiten.
  if( !ParseScript() )
    return;

  ParseCommandEnd();
}


void ParseTclIncrCmd( void )
{
  if( !ParseUnit( "varName" ) )
    return;

  if( !IsVariable( Actual( unit ) ) )
    display.UnknownVariable( Actual( unit ) );

  ParseGeneric( 0, 1 );
}


void ParseTclInfoCmd( void )
{
  //Den Namen des Unterbefehls einlesen.
  if( !ParseUnit() )
    return;

  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Buffer() );

  if( !strcmp( "args", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "body", command ) )
    ParseGeneric( 1, 1 );
    
  else if( !strcmp( "cmdcount", command ) )
    ParseCommandEnd();
    
  else if( !strcmp( "commands", command ) )
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( "complete", command ) )
    ParseGeneric( 1, 1 );
    
  else if( !strcmp( "default", command ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( "exists", command ) )
    ParseGeneric( 1, 1 );
    
  else if( !strcmp( "globals", command ) )
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( "hostname", command ) )
    ParseCommandEnd();
    
  else if( !strcmp( "level", command ) )
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( "library", command ) )
    ParseCommandEnd();

  else if( !strcmp( "loaded", command ) )
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( "locals", command ) )
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( "nameofexecutable", command ) )
    ParseCommandEnd();
    
  else if( !strcmp( "patchlevel", command ) )
    ParseCommandEnd();

  else if( !strcmp( "procs", command ) )
    ParseGeneric( 0, 1 );
    
  else if( !strcmp( "script", command ) )
    ParseCommandEnd();
    
  else if( !strcmp( "sharedlibextension", command ) )
    ParseCommandEnd();
    
  else if( !strcmp( "tclversion", command ) )
    ParseCommandEnd();
    
  else if( !strcmp( "vars", command ) )
    ParseGeneric( 0, 1 );

  else
    display.UnknownOption( command, "info" );
}


void ParseTclInterpCmd( void )
{
  //Den Namen des Unterbefehls einlesen.
  if( !ParseUnit() )
    return;

  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Buffer() );

  if( !strcmp( "alias", command ) )
    ParseGeneric( 2, -1 );

  else if( !strcmp( "aliases", command ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( "create", command ) )
    {
      if( !ParseOptions( "interp_create" ) )
	return;

      ParseGeneric( 0, 1 );
    }

  else if( !strcmp( "delete", command ) )
    ParseGeneric( 0, -1 );

  else if( !strcmp( "eval", command ) )
    ParseGeneric( 2, -1 );

  else if( !strcmp( "exists", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "issafe", command ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( "share", command ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( "slaves", command ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( "target", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "transfer", command ) )
    ParseGeneric( 3, 3 );

  else
    display.UnknownOption( command, "interp" );
}


void ParseTclJoinCmd( void )
{
  ParseGeneric( 1, 2 );
}


void ParseTclLappendCmd( void )
{
  ParseTclAppendCmd();
}


void ParseTclLindexCmd( void )
{
  ParseGeneric( 2, 2 );
}


void ParseTclLinsertCmd( void )
{
  ParseGeneric( 3, -1 );
}


void ParseTclListCmd( void )
{
  ParseGeneric( 0, -1 );
}


void ParseTclLlengthCmd( void )
{
  ParseGeneric( 1, 1 );
}


void ParseTclLoadCmd( void )
{
  ParseGeneric( 1, 3 );
}


void ParseTclLrangeCmd( void )
{
  ParseGeneric( 3, 3 );
}


void ParseTclLreplaceCmd( void )
{
  ParseGeneric( 3, -1 );
}


void ParseTclLsearchCmd( void )
{
  if( !ParseOptions( "lsearch" ) )
    return;

  ParseGeneric( 2, 2 );
}


void ParseTclLsortCmd( void )
{
  if( !ParseOptions( "lsort" ) )
    return;

  ParseGeneric( 1, 1 );
}


void ParseTclOpenCmd( void )
{
  ParseGeneric( 1, 3 );
}


void ParseTclPackageCmd( void )
{
  //Den Namen des Unterbefehls einlesen.
  if( !ParseUnit() )
    return;

  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Buffer() );

  if( !strcmp( "forget", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "ifneeded", command ) )
    ParseGeneric( 2, 3 );

  else if( !strcmp( "names", command ) )
    ParseCommandEnd();

  else if( !strcmp( "provide", command ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( "require", command ) )
    {
      if( !ParseUnit() )
	return;

      if( strcmp( "-exact", Actual( unit ) ) )
	ParseGeneric( 0, 1 );
      else
	ParseGeneric( 1, 2 );
    }

  else if( !strcmp( "unknown", command ) )
    ParseGeneric( 0, 1 );

  else if( !strcmp( "vcompare", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "versions", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "vsatisfies", command ) )
    ParseGeneric( 2, 2 );

  else
    display.UnknownOption( command, "package" );
}


void ParseTclPidCmd( void )
{
  ParseGeneric( 0, 1 );
}


void ParseTclProcCmd( void )
{
  if( !ParseUnit( "procName" ) )
    return;

  char procName[ strlen( Actual( unit ) )+1 ];
  strcpy( procName, Buffer() );

  display.Procedure( procName );

  CCommand *procedure = ParseArgumentList();

  if( procedure == NULL)
    return;

  procedure->SetName( procName );
  procedure->SetCommandID( "PROCEDURE" );
  AddCommand( *procedure );
  actualFile->stack.Push( procedures, procName );

  //Die Argumente fuer den Rumpf sichtbar machen.
  char list[ strlen( procedure->List() )+1 ];
  strcpy( list, Buffer() );
  char *dummy = strtok( list, " " );
  while( dummy != NULL )
    {
      AddVariable( dummy, Environment(), 1, 0 );
      dummy = strtok( NULL, " " );
    }

  //Den Rumpf bearbeiten.
  if( !ParseScript() )
    return;

  CVariable variable;
  //Die Argumente wieder loeschen.
  if( procedure->Min() > 0 )
    {
      strcpy( list, procedure->List() );
      char *dummy = strtok( list, " " );
      while( dummy != NULL )
	{
	  variable.SetName( dummy );
	  variable.SetEnvironment( Environment() );
	  RemoveVariable( variable );
	  dummy = strtok( NULL, " " );
	}
    }

  actualFile->stack.Pop( procedures );
  ParseCommandEnd();

  display.ProcedureEnd( procName );

  delete procedure;
}


void ParseTclPutsCmd( void )
{
  if( !ParseUnit() )
    return;

  if( !strcmp( Actual( unit ), "-nonewline" ) )
    ParseGeneric( 1, 2 );

  else
    ParseGeneric( 0, 1 );
}


void ParseTclPwdCmd( void )
{
  ParseCommandEnd();
}


void ParseTclReadCmd( void )
{
  ParseGeneric( 1, 2 );
}


void ParseTclRegexpCmd( void )
{
  if( !ParseOptions( "regexp" ) )
    return;

  if( !ParseUnit( "exp" ) )
    return;

  if( !ParseUnit( "string" ) )
    return;

  char copy[ unitLength ];

  //Alle Argumente die jetzt noch folgen werden als Variablen sichtbar gemacht.
  while( ParseOptionalUnit() )
    {
      strcpy( copy, Buffer() );
      AddVariable( copy, Environment(), 1, 0 );
    }
}


void ParseTclRegsubCmd( void )
{
  if( !ParseOptions( "regsub" ) )
    return;

  if( !ParseUnit( "exp" ) )
    return;

  if( !ParseUnit( "string" ) )
    return;

  if( !ParseUnit( "subSpec" ) )
    return;

  if( !ParseUnit( "varName" ) )
    return;

  char copy[ strlen( Actual( unit ) )+1 ];
  strcpy( copy, Buffer() );

  //Die Variable noch sichtbar machen.
  AddVariable( copy, Environment(), 1, 0 );
}


void ParseTclRenameCmd( void )
{
  ParseGeneric( 2, 2 );
}


void ParseTclReturnCmd( void )
{
  if( !ParseOptions( "return" ) )
    return;

  if( !ParseOptionalUnit() )
    return;

  ParseCommandEnd();
}


void ParseTclScanCmd( void )
{
  if( !ParseUnit( "string" ) )
    return;

  if( !ParseUnit( "format" ) )
    return;

  if( !ParseUnit( "varName" ) )
    return;

  char copy[ unitLength ];
  strcpy( copy, Buffer() );

  //Die Variable sichtbar machen.
  AddVariable( copy, Environment(), 1, 0 );

  //Jetzt koennen noch optionale Variablennamen folgen.
  while( ParseOptionalUnit() )
    {
      strcpy( copy, Buffer() );
      AddVariable( copy, Environment(), 1, 0 );
    }
}


void ParseTclSeekCmd( void )
{
  ParseGeneric( 2, 3 );
}


void ParseTclSetCmd( void )
{
  //Den Variablennamen einlesen.
  if( !ParseUnit() )
    return;

  char varName[ strlen( Actual( unit ) )+1 ];
  strcpy( varName, Buffer() );

  //Den optionalen Wert einlesen.
  if( !ParseOptionalUnit() )
    return;

  //Wurde ein Wert angegeben, so handelt es sich um eine Variablendefinition.
  AddVariable( varName, Environment(), 1, 0 );

  //Das Befehlsende wird erwartet.
  ParseCommandEnd();
}


void ParseTclSocketCmd( void )
{
  if( IsLastArg() )
    {
      display.Error( "argument expected" );
      return;
    }

  //Zuerst ermitteln, welche Art des Befehls benutzt wird.
  if( !strcmp( Next( unit ), "-server" ) )
    {
      GetUnit();

      if( !ParseUnit( "command" ) )
	return;

      if( !ParseOptions( "socket" ) )
	return;

      if( !ParseUnit( "port" ) )
	return;
    }

  else
    {
      if( !ParseOptions( "socket" ) )
	return;

      if( !ParseUnit( "host" ) )
	return;

      if( !ParseUnit( "port" ) )
	return;
    }

  ParseCommandEnd();
}


void ParseTclSourceCmd( void )
{
  if( !ParseUnit() )
    return;

  ParseCommandEnd();
}


void ParseTclSplitCmd( void )
{
  ParseGeneric( 1, 2 );
}


void ParseTclStringCmd( void )
{
  if( !ParseUnit() )
    return;

  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Buffer() );

  if( !strcmp( "compare", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "first", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "index", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "last", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "length", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "match", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "range", command ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( "tolower", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "toupper", command ) )
    ParseGeneric( 1, 1 );

  else if( !strcmp( "trim", command ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( "trimleft", command ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( "trimright", command ) )
    ParseGeneric( 1, 2 );

  else if( !strcmp( "wordend", command ) )
    ParseGeneric( 2, 2 );

  else if( !strcmp( "wordstart", command ) )
    ParseGeneric( 2, 2 );

  else
    display.UnknownOption( command, "string" );
}


void ParseTclSubstCmd( void )
{
  if( !ParseOptions( "subst" ) )
    return;

  ParseGeneric( 1, 1 );
}


void ParseTclSwitchCmd( void )
{
  if( !ParseOptions( "switch" ) )
    return;

  if( !ParseUnit( "string" ) )
    return;

  //Feststellen, welche Art des Befehls verwendet wird.
  if( !strcmp( NextID( word ), "START_BRACE_GROUP" ) )
    {
      GetWord();

      while( 1 )
	{
	  //Argument 'pattern' einlesen.
	  GetUnit();

	  //Eventueller Kommentar muss extra abgefragt werden.
	  if( !strcmp( ActualID( unit ), "HASH" ) )
	    {
	      ParseComment();
	      continue;
	    }

	  //Ist das Befehlsende erreicht?
	  if( !strcmp( ActualID( unit ), "END_BRACE_GROUP" ) )
	    return;

	  else if( !strcmp( ActualID( unit ), "EOF" ) )
	    {
	      display.Error( "end of file encountered in switch body" );
	      return;
	    }

	  //Den zugehoerigen Rumpf einlesen. Eventuell ist dies aber auch nur
	  //ein Bindestrich.
	  if( !strcmp( Next( word ), "-" ) )
	    GetUnit();

	  else
	    {
	      if( !ParseScript() )
		return;
	    }
	}
    }

  else if( !strcmp( NextID( word ), "START_QUOTE_GROUP" ) )
    {
      GetWord();

      while( 1 )
	{
	  //Argument 'pattern' einlesen.
	  GetUnit();

	  //Eventueller Kommentar muss extra abgefragt werden.
	  if( !strcmp( ActualID( unit ), "HASH" ) )
	    {
	      ParseComment();
	      continue;
	    }

	  //Ist das Befehlsende erreicht?
	  if( !strcmp( ActualID( unit ), "END_QUOTE_GROUP" ) )
	    return;

	  else if( !strcmp( ActualID( unit ), "EOF" ) )
	    {
	      display.Error( "end of file encountered in switch body" );
	      return;
	    }

	  //Den zugehoerigen Rumpf einlesen.
	  if( !ParseScript() )
	    return;
	}
    }

  else
    {
      //Zuerst das Argument 'pattern' einlesen.
      if( !ParseUnit( "pattern" ) )
	return;

      //Dann den zugehoerigen Rumpf einlesen.
      if( !ParseScript() )
	return;

      //Jetzt koennen noch optionale 'paattern'-Rumpf-Paare folgen.
      while( 1 )
	{
	  //Argument 'pattern' einlesen.
	  if( !ParseOptionalUnit() )
	    return;

	  //Den zugehoerigen Rumpf einlesen.
	  if( !ParseScript() )
	    return;
	}
    }
}


void ParseTclTellCmd( void )
{
  ParseGeneric( 1, 1 );
}


void ParseTclTimeCmd( void )
{
  if( !ParseScript() )
    return;

  ParseGeneric( 0, 1 );
}


void ParseTclTraceCmd( void )
{
  if( !ParseUnit() )
    return;

  char command[ strlen( Actual( unit ) )+1 ];
  strcpy( command, Actual( unit ) );

  if( !strcmp( "variable", command ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( "vdelete", command ) )
    ParseGeneric( 3, 3 );

  else if( !strcmp( "vinfo", command ) )
    ParseGeneric( 1, 1 );

  else
    display.UnknownOption( command, "trace" );
}


void ParseTclUnknownCmd( void )
{
  ParseGeneric( 1, -1 );
}


void ParseTclUnsetCmd( void )
{
  if( !ParseUnit() )
    return;

  char copy[ strlen( Actual( unit ) ) + 1 ];
  strcpy( copy, Buffer() );

  CVariable variable( copy, Environment(), 0, 0 );
  RemoveVariable( variable );
}


void ParseTclUnsupported0Cmd( void )
{
  ParseGeneric( 2, 3 );
}


void ParseTclUpdateCmd( void )
{
  ParseGeneric( 0, 1 );
}


void ParseTclUplevelCmd( void )
{
  if( !ParseUnit() )
    return;

  //Pruefen, ob das 'level'-Argument angegeben wurde.
  if( (Actual( unit )[ 0 ] == '#') || (IsDigit( Actual( unit )[ 0 ] )) )
    {
      if( !IsLevel( Actual( unit ) ) )
	{
	  HandleError();
	  return;
	}

      ParseGeneric( 1, -1 );
    }

  else
    ParseGeneric( 0, -1 );
}


void ParseTclUpvarCmd( void )
{
  if( !ParseUnit() )
    return;

  //Pruefen, ob das 'level'-Argument angegeben wurde.
  if( (Actual( unit )[ 0 ] == '#') || (IsDigit( Actual( unit )[ 0 ] )) )
    {
      if( !IsLevel( Actual( unit ) ) )
	{
	  HandleError();
	  return;
	}

      if( !ParseUnit( "otherVar" ) )
	return;
    }

  if( !ParseUnit( "myVar" ) )
    return;

  char copy[ unitLength ];
  strcpy( copy, Buffer() );

  //Die Variable sichtbar machen.
  AddVariable( copy, Environment(), 0, 0 );

  //Es koennen noch optionale 'otherVar'-'myVar'-Paare folgen.
  while( 1 )
    {
      if( !ParseOptionalUnit() )
	return;

      if( !ParseUnit( "myVar" ) )
	return;

      //Die Variable sichtbar machen.
      strcpy( copy, Buffer() );
      AddVariable( copy, Environment(), 0, 0 );
    }
}


void ParseTclVwaitCmd( void )
{
  ParseGeneric( 0, 1 );
}


void ParseTclWhileCmd( void )
{
  if( !ParseUnit() )
    return;

  if( !ParseScript() )
    return;

  ParseCommandEnd();
}
